// SPDX-License-Identifier: Apache-2.0

// Copyright 2019-2022 Fraunhofer AISEC
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
//    $$\      $$\ $$$$$$$$\ $$$$$$$\  $$$$$$\ $$\   $$\  $$$$$$\
//    $$$\    $$$ |$$  _____|$$  __$$\ \_$$  _|$$$\  $$ |$$  __$$\
//    $$$$\  $$$$ |$$ |      $$ |  $$ |  $$ |  $$$$\ $$ |$$ /  $$ |
//    $$\$$\$$ $$ |$$$$$\    $$ |  $$ |  $$ |  $$ $$\$$ |$$$$$$$$ |
//    $$ \$$$  $$ |$$  __|   $$ |  $$ |  $$ |  $$ \$$$$ |$$  __$$ |
//    $$ |\$  /$$ |$$ |      $$ |  $$ |  $$ |  $$ |\$$$ |$$ |  $$ |
//    $$ | \_/ $$ |$$$$$$$$\ $$$$$$$  |$$$$$$\ $$ | \$$ |$$ |  $$ |
//    \__|     \__|\________|\_______/ \______|\__|  \__|\__|  \__|
//
// This file is part of the MEDINA Framework.

package main

import (
	"collector"
	"context"
	"encoding/json"
	"fmt"
	"net/http"
	"os"

	"clouditor.io/clouditor/api/discovery"
	"clouditor.io/clouditor/persistence"
	"clouditor.io/clouditor/persistence/gorm"
	server_clouditor "clouditor.io/clouditor/server"
	"clouditor.io/clouditor/server/rest"
	service_discovery "clouditor.io/clouditor/service/discovery"
	"clouditor.io/clouditor/voc"

	structpb "github.com/golang/protobuf/ptypes/struct"
	"github.com/sirupsen/logrus"
	"golang.org/x/oauth2/clientcredentials"
	"google.golang.org/grpc"
)

var (
	log              *logrus.Entry
	server           *grpc.Server
	discoveryService discovery.DiscoveryServer
	storage          persistence.Storage

	assessmentUrl   string
	jwksURL         string
	config          clientcredentials.Config
	cloudServiceID  string
	resourceGroupId string
	dbHost          string
	dbUser          string
	dbPassword      string

	// provider sets the cloud service provider an evidence collector instance (pod) should gather resources from
	provider string
)

const (
	grpcPort             = 9091
	httpPort             = 8081
	AssessmentUrl        = "ASSESSMENT_URL"
	DefaultAssessmentURL = "localhost:9092"
	OAuth2ClientID       = "OAUTH2_CLIENT_ID"
	OAuth2ClientSecret   = "OAUTH2_CLIENT_SECRET"
	OAuth2TokenURL       = "OAUTH2_TOKEN_URL"
	JWKSURL              = "AUTH_JWKS_URL"
	CloudServiceID       = "CLOUD_SERVICE_ID"
	ResourceGroupId      = "RESOURCE_GROUP_ID"
	PostgresHost         = "POSTGRES_HOST"
	PostgresUser         = "POSTGRES_USER"
	PostgresPassword     = "POSTGRES_PASSWORD"
	Provider             = "PROVIDER"
)

func toStruct(r voc.IsCloudResource) (s *structpb.Value) {
	var (
		b   []byte
		err error
	)

	s = new(structpb.Value)

	// this is probably not the fastest approach, but this
	// way, no extra libraries are needed and no extra struct tags
	// except `json` are required. there is also no significant
	// speed increase in marshaling the whole resource list, because
	// we first need to build it out of the map anyway
	if b, err = json.Marshal(r); err != nil {
		return
	}
	if err = json.Unmarshal(b, &s); err != nil {
		return
	}
	return
}

func init() {
	log = logrus.WithField("component", "grpc")
	logrus.SetLevel(logrus.DebugLevel)

	// Get URL for Clouditor Assessment from env var
	var ok bool
	assessmentUrl, ok = os.LookupEnv(AssessmentUrl)
	if !ok {
		// If no environment variable available set to default
		assessmentUrl = DefaultAssessmentURL
	}

	// Get the JWKS URL of our auth server
	jwksURL, _ = os.LookupEnv(JWKSURL)

	cloudServiceID, ok = os.LookupEnv(CloudServiceID)
	if !ok {
		cloudServiceID = discovery.DefaultCloudServiceID
	}

	resourceGroupId, ok = os.LookupEnv(ResourceGroupId)
	if !ok {
		resourceGroupId = ""
	}

	// Get the OAuth credentials
	config = clientcredentials.Config{}
	config.ClientID, _ = os.LookupEnv(OAuth2ClientID)
	config.ClientSecret, _ = os.LookupEnv(OAuth2ClientSecret)
	config.TokenURL, _ = os.LookupEnv(OAuth2TokenURL)

	// Get DB credentials
	dbHost, _ = os.LookupEnv(PostgresHost)
	dbUser, _ = os.LookupEnv(PostgresUser)
	dbPassword, _ = os.LookupEnv(PostgresPassword)
}

func main() {
	var (
		textFormatter = logrus.TextFormatter{ForceColors: false, FullTimestamp: true}
		err           error
	)
	log.Logger.Formatter = &textFormatter

	fmt.Printf(`
	$$\      $$\ $$$$$$$$\ $$$$$$$\  $$$$$$\ $$\   $$\  $$$$$$\                                                           
	$$$\    $$$ |$$  _____|$$  __$$\ \_$$  _|$$$\  $$ |$$  __$$\                                                          
	$$$$\  $$$$ |$$ |      $$ |  $$ |  $$ |  $$$$\ $$ |$$ /  $$ |                                                         
	$$\$$\$$ $$ |$$$$$\    $$ |  $$ |  $$ |  $$ $$\$$ |$$$$$$$$ |                                                         
	$$ \$$$  $$ |$$  __|   $$ |  $$ |  $$ |  $$ \$$$$ |$$  __$$ |                                                         
	$$ |\$  /$$ |$$ |      $$ |  $$ |  $$ |  $$ |\$$$ |$$ |  $$ |                                                         
	$$ | \_/ $$ |$$$$$$$$\ $$$$$$$  |$$$$$$\ $$ | \$$ |$$ |  $$ |                                                         
	\__|     \__|\________|\_______/ \______|\__|  \__|\__|  \__| 

	Evidence Collector 1.1 based on Clouditor framework %s
	`, collector.ClouditorVersion())

	log.Infof("Security Assessment URL is set to %s", assessmentUrl)

	storage, err = gorm.NewStorage(gorm.WithPostgres(dbHost, 5432, dbUser, dbPassword, "postgres", ""))
	if err != nil {
		log.Errorf("Could not connect to storage (yet): %v", err)
		os.Exit(1)
	}

	// Get provider (one is enough because 1 cloud service == 1 provider)
	provider = os.Getenv(Provider)
	if provider == "" {
		log.Errorf("Exiting. No provider specified: %v", err)
		os.Exit(1)
	}

	discoveryService = service_discovery.NewService(
		service_discovery.WithOAuth2Authorizer(&config),
		service_discovery.WithProviders([]string{provider}),
		service_discovery.WithAssessmentAddress(assessmentUrl),
		service_discovery.WithCloudServiceID(cloudServiceID),
		service_discovery.WithStorage(storage),
	)

	// Comment in if evidence collector should collect the evidences from the CSPs. Otherwise, example evidences are used.
	// Start evidence collector (possibly scoped to the resourcegroup)
	if resourceGroupId == "" {
		_, err = discoveryService.Start(context.Background(), &discovery.StartDiscoveryRequest{})
	} else {
		_, err = discoveryService.Start(context.Background(), &discovery.StartDiscoveryRequest{
			ResourceGroup: &resourceGroupId,
		})
	}

	if err != nil {
		log.Errorf("could not collect evidences: %v", err)
	}

	log.Infof("Starting gRPC endpoint on :%d", grpcPort)

	// Add grpc opts
	grpcOpts := []grpc.ServerOption{
		// Add max grpc message sizes
		grpc.MaxRecvMsgSize(1024 * 1024 * 20),
		grpc.MaxSendMsgSize(1024 * 1024 * 20)}

	// Start the gRPC server
	_, server, err = server_clouditor.StartGRPCServer(
		fmt.Sprintf("0.0.0.0:%d", grpcPort),
		server_clouditor.WithJWKS(jwksURL),
		server_clouditor.WithDiscovery(discoveryService),
		server_clouditor.WithReflection(),
		server_clouditor.WithAdditionalGRPCOpts(grpcOpts),
	)
	if err != nil {
		log.Errorf("Failed to serve gRPC endpoint: %s", err)
		return
	}

	// Start the gRPC-HTTP gateway
	err = rest.RunServer(context.Background(),
		grpcPort,
		httpPort,
	)
	if err != nil && err != http.ErrServerClosed {
		log.Errorf("failed to serve gRPC-HTTP gateway: %v", err)
		return
	}

	log.Infof("Stopping gRPC endpoint")
	server.GracefulStop()
}
